#
# Basic Markdown parser in Tcl
#

package require snit

snit::type markdown {
  #
  # Options
  #
  
  option -debug 0
  
  #
  # Variables
  #
  variable rowNo       ;# Current character indexes row number in content
  variable colNo       ;# Current character indexes number in content
  variable idx         ;# Current character index in content
  
  variable content     ;# Content we are parsing
  variable contentSize ;# String size of the content
  variable tokens      ;# Parsed markdown tokens
  variable tagStack    ;# Current tag stack
  
  variable ch          ;# Current character
  variable txt         ;# Text accumulator
  
  #
  # Elements:
  #  1. Name
  #  2. Push onto the tag stack?
  #  3. Auto terminate when?
  #  4. Creates what type of output?
  #
  
  variable tagTypes {
    { NL            0 NONE NONE }
    { PARA          1 PARA BLOCK }
    { HEADING_1     1 LINE BLOCK }
    { HEADING_2     1 LINE BLOCK }
    { HEADING_3     1 LINE BLOCK }
    { HEADING_4     1 LINE BLOCK }
    { HEADING_5     1 LINE BLOCK }
    { HEADING_6     1 LINE BLOCK }
    { EMPHASIS      1 PARA INLINE }
    { STRONG        1 PARA INLINE }
    { CODE          1 PARA INLINE }
    { ANCHOR        1 PARA INLINE }
    { TEXT          0 NONE INLINE }
  }
  
  constructor {args} {
    $self configurelist $args
    
    set idx -1   ;# BOF
    set rowNo 1
    set colNo 0
    set contentSize 0
    set tokens {}
    set tagStack {}
    set txt ""
  }
  
  #
  # Parse `s` as Markdown text
  #
  method parse {s} {
    regsub -all "\r" "" $s
    set s [string trim $s]
    
    set content $s
    set contentSize [string length $content]
    
    while {[$self nextChar]} {
      if {$options(-debug)} {
        puts "$rowNo:$colNo ch=$ch ($tagStack)"
      }
      
      if {$ch=="\n"} {
        # Loop through our `tagStack` and determine if we need to close any LINE tags.
        while {$tagStack!={}} {
          if {[lindex $tagStack end 2]=="LINE"} {
            $self addMarkup END_[lindex $tagStack end 0] -add 0
            $self popTag
            continue
          }
          
          break
        }
        
        incr rowNo
        set colNo 0
        
        $self addMarkup NL
      } elseif {$colNo==1 && $ch=="#"} {
        set headingLevel 1
        while {[$self nextChar] && $ch=="#"} {
          incr headingLevel
        }
        
        # Put back the character we read that was not #
        $self putbackChar
        
        $self addMarkup HEADING_$headingLevel
      } elseif {($ch=="*" && [$self lookAhead]=="*") || ($ch=="_" && [$self lookAhead]=="_")} {
        $self nextChar ;# Read the next *
        
        if {[$self isTopTag STRONG]} {
          $self addMarkup END_STRONG -add 0
          $self popTag
        } else {
          $self addMarkup STRONG
        }
      } elseif {$ch=="*" || $ch=="_"} {
        if {[$self isTopTag EMPHASIS]} {
          $self addMarkup END_EMPHASIS -add 0
          $self popTag
        } else {
          $self addMarkup EMPHASIS
        }
      } elseif {$ch=="`"} {
        if {[$self isTopTag CODE]} {
          $self addMarkup END_CODE -add 0
          $self popTag
        } else {
          $self addMarkup CODE
        }
      } elseif {$ch=={[}} {
        $self addMarkup ANCHOR
      } elseif {$ch=={]} && [$self isTopTag ANCHOR]} {
        lset [myvar tokens] end 1 $txt
        set txt ""
      } elseif {$ch=={(} && [$self isTopTag ANCHOR]} {
      } elseif {$ch=={)} && [$self isTopTag ANCHOR]} {
        set urlComponents [split $txt]
        lset [myvar tokens] end 2 [list -url [lindex $urlComponents 0] \
          -title [join [lrange $urlComponents 1 end]]]        
        set txt ""
        $self popTag
      } else {
        if {$ch=="\\"} {
          if {[$self nextChar]} {
            switch -- $ch {
              \\ {}
              \` {}
              \* {}
              \_ {}
              \{ {}
              \} {}
              \[ {}
              \] {}
              \( {}
              \) {}
              \# {}
              \+ {}
              \- {}
              \. {}
              \! {}
              default {
                $self putbackChar
              }
            }
          }
        }
        
        $self addText $ch
      }
    }
    
    $self addTextToken?
    $self fixupTokens
    
    return $tokens
  }
  
  #
  # Fix up tokens such as adding paragraphs, stripping spaces where necessary.
  #
  method fixupTokens {} {
    for {set x 0} {$x < [llength $tokens]} {incr x} {
      set lastTag [$self findTag [lindex $tokens $x-1 0]]
      set thisTag [$self findTag [lindex $tokens $x 0]]
      set nextTag [$self findTag [lindex $tokens $x+1 0]]
      
      if {[lindex $lastTag 0]=="TEXT" && [lindex $thisTag 0]=="NL" && [lindex $nextTag 0]=="TEXT"} {
        set lastLineText [lindex $tokens $x-1 1]
        set nextLineText [lindex $tokens $x+1 1]
        if {[regexp {^=+$} $nextLineText]} {
          set tokens [lreplace $tokens $x-1 $x+1 \
              [list HEADING_1 {} {}] \
              [list TEXT $lastLineText {}] \
              [list END_HEADING_1 {} {}] \
            ]
        } elseif {[regexp {^-+$} $nextLineText]} {
          set tokens [lreplace $tokens $x-1 $x+1 \
              [list HEADING_2 {} {}] \
              [list TEXT $lastLineText {}] \
              [list END_HEADING_2 {} {}] \
            ]
        }
        continue
      }
      if {[lindex $thisTag 0]=="NL" && [lindex $lastTag 0]=="NL" && [lindex $nextTag 3]=="INLINE"} {
        set tokens [lreplace $tokens $x-1 $x [list PARA {}]]
        incr x -2
        continue
      } elseif {[lindex $thisTag 0]=="NL" && [lindex $nextTag 0]=="NL" && [lindex $lastTag 3]=="INLINE"} {
        set tokens [lreplace $tokens $x $x+1 [list END_PARA {}]]
        incr x -2
        continue
      }
      
      if {[lindex $lastTag 3]=="BLOCK" && [lindex $thisTag 0]=="TEXT"} {
        lset [myvar tokens] [list $x 1] [string trimleft [lindex $tokens $x 1]]
      }
    }
    
    # If the last tag is an INLINE, then end with a paragraph.
    set lastTag [$self findTag [lindex $tokens end 0]]
    if {[lindex $lastTag 3]=="INLINE"} {
      lappend tokens [list END_PARA {}]
    }
  }
  
  #
  # Read the next character in the content. Set `ch` and return `1` on success or `0` on failure.
  #
  method nextChar {} {
    if {$idx>=$contentSize} {
      return 0
    }
    
    incr idx
    incr colNo
    
    set ch [string index $content $idx]
    
    return 1
  }
  
  #
  # Puts a character back onto the content by backing up `idx`.
  #
  method putbackChar {} {
    if {$idx<=0} {
      return 0
    }
    
    incr idx -1
    incr colNo -1
    
    if {$colNo<1} {
      error "character put back onto content stack went past BOL which is not yet supported."
    }
    
    set ch [string index $content $idx]
    
    return 1
  }
  
  #
  # Look ahead `amount` characters. Returns up to `amount` characters further than `idx`. The length
  # of the string returned may be limited if you have reached the end of `content`.
  #
  method lookAhead {{amount 1}} {
    if {[expr $idx + $amount] >= $contentSize} {
      set amount [expr $contentSize - $idx - 1]
    }
    
    if {$amount==0} {
      return 0
    }
    
    return [string range $content [expr $idx + 1] [expr $idx + $amount]]
  }
  
  #
  # Look behind `amount` characters. Returns up to `amount` characters before `idx`. The length of
  # the string returned may be limited if you have reached the beginning of `content`.
  #
  method lookBehind {{amount 1}} {
    if {[expr $idx - $amount] < 0} {
      set amount $idx
    }
    
    return [string range [expr $idx - $amount] [expr $idx - 1]]
  }
  
  #
  # Find and return the tag `t` from the `tagTypes` list.
  #
  method findTag {ft} {
    foreach t $tagTypes {
      if {$ft==[lindex $t 0]} {
        return $t
      }
    }
    
    return 0
  }
  
  #
  # Add markup onto the `tokens` variable. `token` is required. `data` may be any text associated
  # with the markup. `args` is placed onto the token stack verbatium and can be used for optional
  # markup modifiers.
  #
  method addMarkup {token args} {
    set data {}
    set add 1
    set try_addtext 1
    
    foreach {n v} $args {
      switch -- $n {
        -data { set data $v }
        -add  { set add $v }
        -addtext { set try_addtext $v }
      }
    }
    
    if {$try_addtext} {
      $self addTextToken?
    }

    if {$add} {    
      # Do we need to push this tag onto the `tagStack`?
      set t [$self findTag $token]
      if {$t==0} {
        error "Invalid token added to the markup stack: `$token`"
      }
  
      if {[lindex $t 1]==1} {
        $self pushTag $t
      }
    }
    
    lappend [myvar tokens] [list $token $data $args]
  }
  
  #
  # Add text onto the `txt` accumulator variable. Anything put into `txt` will be automatically
  # added to the `tokens` list before any new tag is added.
  #
  method addText {t} {
    append [myvar txt] $t
  }
  
  #
  # Adds `txt` onto the `token` list if `txt` is not empty. It also resets `txt` to an empty string.
  #
  method addTextToken? {} {
    if {$txt==""} {
      return 0
    }
    
    #$self addMarkup TEXT -data $txt -addtext 0
    lappend [myvar tokens] [list TEXT $txt {}]
    set txt ""
    
    return 1
  }
  
  #
  # Push a tag onto the `tagStack`. This tracks open/closed tags. For instance, when finding the
  # beginning of a header, HEADING_1 is pushed onto the tag stack. Since HEADING_1 is defined as
  # a LINE tag, when we reach the end of the line, if a HEADING_1 is still on the `tagStack` then
  # an END_HEADING_1 is added to `tokens` and HEADING_1 is removed from the `tagStack`.
  #
  method pushTag {tag} {
    lappend [myvar tagStack] $tag
  }
  
  #
  # Determine if the top tag on `tagStack` is the type `tagType`.
  #
  method isTopTag {tagType} {
    if {$tagStack=={}} {
      return 0
    }
    
    if {[lindex $tagStack end 0]==$tagType} {
      return 1
    }
    
    return 0
  }
  
  #
  # Pop a tag off of the `tagStack`. Returns the tag on the top of the stack or 0 if the stack was
  # empty.
  #
  method popTag {} {
    if {$tagStack=={}} {
      return 0
    }
    
    set tt [lindex $tagStack end]
    set tagStack [lrange $tagStack 0 end-1]
    
    return $tt
  }
}

set example {
# This is an example

Hello _World_! How *are* you __doing__ today?

## This is a heading 2
### Heading 3
#### Heading 4
##### Heading 5
###### Heading 6

Please **refer** to the function `info` for more information.

This is another heading 1
=========================

How are you [doing today](http://google.com?q=today)?
This is [pretty](http://google.com?q=pretty Google for Pretty)

This is another heading 2
--------------------

How are you doing today? These should all be escaped.
\*hello!\*
\`what?\`
\_who?\_
}

proc markdown_to_html {fh tokens {head "<html><body>"} {foot "</body></html>"}} {
  puts $fh $head
  
  foreach t $tokens {
    set tn [lindex $t 0]
    set tv [lindex $t 1]
    set ta [lindex $t 2]
    
    switch -- $tn {
      NL            { puts -nonewline $fh " " }
      PARA          { puts $fh "<p>" }
      END_PARA      { puts $fh "\n</p>" }
      HEADING_1     { puts -nonewline $fh "<h1>" }
      END_HEADING_1 { puts $fh "</h1>" }
      HEADING_2     { puts -nonewline $fh "<h2>" }
      END_HEADING_2 { puts $fh "</h2>" }
      HEADING_3     { puts -nonewline $fh "<h3>" }
      END_HEADING_3 { puts $fh "</h3>" }
      HEADING_4     { puts -nonewline $fh "<h4>" }
      END_HEADING_4 { puts $fh "</h4>" }
      HEADING_5     { puts -nonewline $fh "<h5>" }
      END_HEADING_5 { puts $fh "</h5>" }
      HEADING_6     { puts -nonewline $fh "<h6>" }
      END_HEADING_6 { puts $fh "</h6>" }
      EMPHASIS      { puts -nonewline $fh "<i>" }
      END_EMPHASIS  { puts -nonewline $fh "</i>" }
      STRONG        { puts -nonewline $fh "<strong>" }
      END_STRONG    { puts -nonewline $fh "</strong>" }
      CODE          { puts -nonewline $fh "<tt>" }
      END_CODE      { puts -nonewline $fh "</tt>" }
      TEXT          { puts -nonewline $fh $tv }
      ANCHOR        {
        set attrs(-url) ""
        set attrs(-title) ""
        array set attrs $ta
        
        if {$attrs(-title)!=""} {
          set attrs(-title) " title=\"$attrs(-title)\""
        }
        
        puts -nonewline $fh "<a$attrs(-title) href=\"$attrs(-url)\">$tv</a>"
      }
    }
  }
  
  puts $fh $foot
}

set debug [expr [lsearch $argv -debug]!=-1]

::markdown md -debug $debug
set tokens [md parse $example]
if {$debug} {
  foreach t $tokens {
    puts $t
  }
}

set fh [open test.html w]
markdown_to_html $fh $tokens
close $fh
