#****
# [manpage_begin user n 0.5]
# [moddesc {User handling for Mailroom}]
# [copyright {Copyright 2009 Jeremy Cowgar. All Rights Reserved.}]
# [category {User Handling}]
# [keywords mailroom user]
# [description]
# Provides procedures for maintaining the user database in Mailroom.
#

namespace eval ::mailroom::user {
  #****
  # [section Commands]
  #
  # [list_begin definitions]
  #
  
  #****
  # [call [cmd {::mailroom::user new_from}] [arg addr] [opt [arg time]]]
  #
  # Create/Log activity from the user(s) contained in [arg addr]. If [arg time] is specified, the `recent`
  # attribute of the user is updated to reflect the value contained in [arg time].
  #
  proc new_from {addr {time 0}} {
    return [_new $addr $time incoming]
  }
  
  #****
  # [call [cmd {::mailroom::user new_from}] [arg addr] [opt [arg time]]]
  #
  # Create/Log activity to the user(s) contained in [arg addr]. If [arg time] is specified, the `recent`
  # attribute of the user is updated to reflect the value contained in [arg time].
  #
  proc new_to {addresses args} {
    set toTime 0
    
    foreach {arg val} $args {
      switch -- $arg {
        -time { set toTime $val }
      }
    }
    
    return [_new $addresses $toTime outgoing]
  }
  
  #****
  # [call [cmd {::mailroom::user inc}] [arg userid] [arg field] [opt [arg count]]]
  #
  # Increment user [arg userid]'s [arg field] counter by [arg count] which defaults to 1.
  #
  proc inc {userid field {count 1}} {
    set sql "UPDATE user SET $field=$field+$count WHERE userid=\$userid"
    ::mailroom::db eval $sql
  }
  
  #****
  # [call [cmd {::mailroom::user deincrement}] \
  # [opt "[option -id] [arg user-id]"] \
  # [opt "[option -addresses] [arg addresses]"] \
  # [opt "[option -from] [arg boolean]"] \
  # [opt "[option -to] [arg boolean]"] \
  # ]
  #
  # Deincrement a users to/from counters.
  #
  proc deincrement {args} {
    set id 0
    set addresses {}
    set from 0
    set to 0
    
    foreach {a v} $args {
      switch -- $a {
        -id        { set id $v }
        -addresses { set addresses $v }
        -from      { set from $v }
        -to        { set to $v }
        default {
          error "Invalid argument `$a` given to deincrement. Must -id, -addresses, -from or -to"
        }
      }
    }
    
    if {$to == 0 && $from == 0} {
      error "No operation given to deincrement. You must specify either `-to` or `-from`."
    }
    
    if {$id == 0 && $addresses == {}} {
      error "No user given to deincrement. You must specify either `-id` or `-addresses`."
    }
    
    ::mailroom::db transaction {
      foreach addr [mime::parseaddress $addresses] {
        unset -nocomplain x
        array set x $addr
        if {$to} {
          ::mailroom::db eval {UPDATE user SET to_count=to_count-1 WHERE addr=$x(address)}
        }
        if {$from} {
          ::mailroom::db eval {UPDATE user SET from_count=from_count-1 WHERE addr=$x(address)}
        }
      }
    }
    
    if {$id > 0} {
      if {$to} {
        ::mailroom::db eval {UPDATE user SET to_count=to_count-1 WHERE userid=$id}
      }
      if {$from} {
        ::mailroom::db eval {UPDATE user SET from_count=from_count-1 WHERE userid=$id}
      }
    }
  }

  proc getProper {addr} {
    array set x [lindex [mime::parseaddress $addr] 0]
	return [lindex [::mailroom::db eval {
        SELECT fulladdr FROM user WHERE address=$x(address)
    }] 0]
  }
  
  #
  # Call a user function, if we have user callbacks and the callback exists.
  #
  proc callback {name args} {
    if {$::mailroom::has_user_callbacks == 0} {
      return 1
    }
    
    set name "::user_$cb"
    if {[info procs $name] != ""} {
      if {[catch {uplevel 1 $name {*}$args} result]} {
        tk_messageBox -title "User Callback Error" -message "User Callback Error" \
          -message "$name failed:\n$result"
      }
    }
  }
  
  #####
  #
  # Supporting Functions
  #
  
  #
  #
  # Create/Log activity from/to the given user
  #
  proc _new {addresses time dir} {
    log::log debug "New address: $addresses at $time direction $dir"

    if {$dir == "incoming"} {
      set field from_count
    } else {
      set field to_count
    }
    
    if {$time == 0} {
      set time [::mailroom::julianday]
    }
    
    set userId {}
    set incrSql "UPDATE user SET $field=$field+1 WHERE userid=\$x(userid)"
    
    ::mailroom::db transaction {
      foreach addr [mime::parseaddress $addresses] {
        unset -nocomplain x
        array set x $addr

        if {[regsub {^"'(.*)'" <} $x(proper) {"\1" <} proper2]} {
          set x(proper) $proper2
        }
        
        ::mailroom::db eval {
            SELECT userid, fulladdr AS oldproper, address AS oldaddr, recent, edited 
	    FROM user WHERE address=$x(address)
        } x break
        
        if {![info exists x(userid)]} {
          ::mailroom::db eval {
            INSERT INTO user (address, domain, fulladdr, recent)
            VALUES ($x(address), $x(domain), $x(proper), $time)
          }
          
          set x(recent) $time
          set x(userid) [::mailroom::db last_insert_rowid]
          set x(oldproper) $x(proper)
        }
        
        ::mailroom::db eval $incrSql
        
        if {$dir == "outgoing"} {
	    if {$x(oldproper) != $x(proper)} {
		set fulladdr $x(proper)
		::mailroom::db eval {
		    UPDATE user SET recent=$time, fulladdr=$fulladdr, edited=1
		    WHERE userid=$x(userid)
		}
	    } else {
		::mailroom::db eval {
		    UPDATE user SET recent=$time
		    WHERE userid=$x(userid)
		}
	    }
        } elseif {$x(oldproper) == $x(address) && $x(proper) != $x(address)} {
	    ::mailroom::db eval {
		UPDATE user SET recent=$time, fulladdr=$fulladdr
		WHERE userid=$x(userid)
	    }
	} else {
	    ::mailroom::db eval {
		UPDATE user SET recent=$time
		WHERE userid=$x(userid)
	    }
        }
        
        set userId $x(userid)
      }
    }
    
    return $userId
  }
    
  #****
  # [list_end]
  #
  # [section Examples]
  #
  # [section Todo]
  #
  # [section {Bugs, Ideas, Feedback}]
  # Provide feedback to Jeremy Cowgar <jeremy@cowgar.com>
  
  namespace export new_from new_to inc callback
  namespace ensemble create
}

#****
# [manpage_end]
